@@ -1,10 +1,10 @@
 package Net::DNS::Resolver::Base;
 
 #
-# $Id: Base.pm 1422 2015-10-27 16:01:20Z willem $
+# $Id: Base.pm 1431 2015-12-01 21:18:33Z willem $
 #
 use vars qw($VERSION);
-$VERSION = (qw$LastChangedRevision: 1422 $)[1];
+$VERSION = (qw$LastChangedRevision: 1431 $)[1];
 
 
 use strict;
@@ -37,17 +37,17 @@ use constant PACKETSZ => 512;
 # Olaf Kolkman, RIPE NCC, December 2003.
 
 
-use constant USE_SOCKET => scalar eval {
+use constant USE_SOCKET => defined eval {
 	require IO::Select;
 	require IO::Socket;
 	import IO::Socket;
 };
 
-use constant USE_SOCKET_IP => scalar eval { require IO::Socket::IP; };
+use constant USE_SOCKET_IP => defined eval { require IO::Socket::IP; };
 
-use constant USE_SOCKET_INET => scalar eval { require IO::Socket::INET; };
+use constant USE_SOCKET_INET => defined eval { require IO::Socket::INET; };
 
-use constant USE_SOCKET_INET6 => scalar eval { require IO::Socket::INET6; };
+use constant USE_SOCKET_INET6 => defined eval { require IO::Socket::INET6; };
 
 use constant IPv4 => USE_SOCKET_IP || USE_SOCKET_INET;
 use constant IPv6 => USE_SOCKET_IP || USE_SOCKET_INET6;
@@ -57,7 +57,7 @@ use constant IPv6 => USE_SOCKET_IP || USE_SOCKET_INET6;
 use constant SOCKS => scalar eval { require Config; $Config::Config{usesocks}; };
 
 
-use constant UTIL => scalar eval { require Scalar::Util; };
+use constant UTIL => defined eval { require Scalar::Util; };
 
 sub _tainted { UTIL ? Scalar::Util::tainted(shift) : undef }
 
@@ -80,7 +80,6 @@ sub _untaint {
 		retrans		=> 5,
 		retry		=> 4,
 		usevc		=> ( SOCKS ? 1 : 0 ),
-		stayopen	=> 0,
 		igntc		=> 0,
 		recurse		=> 1,
 		defnames	=> 1,
@@ -100,13 +99,6 @@ sub _untaint {
 		force_v4	=> 0,	# only relevant if IPv6 is supported
 		force_v6	=> 0,	#
 		prefer_v6	=> 0,	# prefer v6, otherwise prefer v4
-		ignqrid         => 0,   # normally packets with non-matching ID
-					# or with the qr bit on are thrown away,
-					# but with 'ignqrid' these packets
-					# are accepted.
-					# USE WITH CARE, YOU ARE VULNERABLE TO
-					# SPOOFING IF SET.
-					# This may be a temporary feature
 		},
 			__PACKAGE__;
 
@@ -127,7 +119,6 @@ my %public_attr = map { $_ => $_ } qw(
 		retrans
 		retry
 		usevc
-		stayopen
 		igntc
 		recurse
 		defnames
@@ -142,6 +133,9 @@ my %public_attr = map { $_ => $_ } qw(
 		cdflag
 		prefer_v4
 		prefer_v6
+
+		ignqrid
+		stayopen
 		);
 
 
@@ -267,23 +261,24 @@ sub string {
 	my $self = shift;
 	$self = $self->_defaults unless ref($self);
 
-	my $IP6line = "prefer_v6\t= $self->{prefer_v6}\tforce_v6    = $self->{force_v6}";
+	my $IP6line = "prefer_v6\t= $self->{prefer_v6}\tforce_v6\t= $self->{force_v6}";
 	my $IP6conf = IPv6 ? $IP6line : '(no IPv6 transport)';
 	my @nslist  = $self->nameservers();
 	my $domain  = $self->domain;
 	return <<END;
 ;; RESOLVER state:
-;;  domain	= $domain
-;;  searchlist	= @{$self->{searchlist}}
-;;  nameservers = @nslist
-;;  port	= $self->{port}
-;;  srcport	= $self->{srcport}	srcaddr	    = $self->{srcaddr}
-;;  tcp_timeout = $self->{tcp_timeout}	udp_timeout = $self->{udp_timeout}
-;;  retrans	= $self->{retrans}	retry	    = $self->{retry}
-;;  defnames	= $self->{defnames}	dnsrch	    = $self->{dnsrch}
-;;  recurse	= $self->{recurse}	usevc	    = $self->{usevc}
-;;  debug	= $self->{debug}	force_v4    = $self->{force_v4}
-;;  $IP6conf
+;; domain	= $domain
+;; searchlist	= @{$self->{searchlist}}
+;; nameservers	= @nslist
+;; defnames	= $self->{defnames}	dnsrch		= $self->{dnsrch}
+;; retrans	= $self->{retrans}	retry		= $self->{retry}
+;; recurse	= $self->{recurse}	igntc		= $self->{igntc}
+;; usevc	= $self->{usevc}	port		= $self->{port}
+;; srcaddr	= $self->{srcaddr}	srcport		= $self->{srcport}
+;; tcp_timeout	= $self->{tcp_timeout}	persistent_tcp	= $self->{persistent_tcp}
+;; udp_timeout	= $self->{udp_timeout}	persistent_udp	= $self->{persistent_udp}
+;; debug	= $self->{debug}	force_v4	= $self->{force_v4}
+;; $IP6conf
 END
 
 }
@@ -314,7 +309,7 @@ sub nameservers {
 	$self = $self->_defaults unless ref($self);
 
 	my ( @ipv4, @ipv6 );
-	foreach my $ns (@_) {
+	foreach my $ns ( grep defined, @_ ) {
 		do { push @ipv6, $ns; next } if _ip_is_ipv6($ns);
 		do { push @ipv4, $ns; next } if _ip_is_ipv4($ns);
 
@@ -375,11 +370,10 @@ sub _cname_addr {
 	my $names = shift;
 
 	map $names->{$_->qname}++, $packet->question;
+	map $names->{$_->cname}++, grep $_->can('cname'), $packet->answer;
 
-	foreach my $rr ( $packet->answer ) {
-		next unless $names->{$rr->name};
+	foreach my $rr ( grep $names->{$_->name}, $packet->answer ) {
 		push( @addr, $rr->address ) if $rr->can('address');
-		$names->{$rr->cname}++ if $rr->can('cname');
 	}
 
 	return @addr;
@@ -459,23 +453,16 @@ sub send {
 	my $packet	= $self->_make_query_packet(@_);
 	my $packet_data = $packet->data;
 
-	my $ans;
+	return $self->_send_tcp( $packet, $packet_data )
+			if $self->{usevc} || length $packet_data > $self->_packetsz;
 
-	if ( $self->{usevc} || length $packet_data > $self->_packetsz ) {
+	my $ans = $self->_send_udp( $packet, $packet_data );
 
-		$ans = $self->_send_tcp( $packet, $packet_data );
+	return $ans if $self->{igntc};
+	return $ans unless $ans && $ans->header->tc;
 
-	} else {
-		$ans = $self->_send_udp( $packet, $packet_data );
-		return $ans if $self->{igntc};
-
-		if ( $ans && $ans->header->tc ) {
-			$self->_diag('packet truncated: retrying using TCP');
-			$ans = $self->_send_tcp( $packet, $packet_data );
-		}
-	}
-
-	return $ans;
+	$self->_diag('packet truncated: retrying using TCP');
+	$self->_send_tcp( $packet, $packet_data );
 }
 
 
@@ -493,7 +480,7 @@ sub _send_tcp {
 		$self->_diag( 'tcp send:', $length, 'bytes' );
 
 		unless ( $sock->send( pack 'n a*', $length, $packet_data ) ) {
-			$self->_diag( 'tcp send:', $self->errorstring($!) );
+			$self->_diag( $self->errorstring($!) );
 			next;
 		}
 
@@ -570,15 +557,15 @@ sub _send_udp {
 		push @ns, [$socket, $ip, $dst_sockaddr];
 	}
 
-	my $retrans = $self->{retrans};
-	my $retry   = $self->{retry};
+	my $retrans = $self->{retrans} || 1;
+	my $retry   = $self->{retry}   || 1;
 	my $servers = scalar(@ns);
 	my $timeout = $servers ? do { no integer; $retrans / $servers } : 0;
 
 	my $lastanswer;
 
 	# Perform each round of retries.
-RETRY: for ( 0 .. $retry ) {
+RETRY: for ( 1 .. $retry ) {					# assumed to be a small number
 
 		# Try each nameserver.
 NAMESERVER: foreach my $ns (@ns) {
@@ -588,7 +575,7 @@ NAMESERVER: foreach my $ns (@ns) {
 			$self->_diag("udp send [$ip]:$port");
 
 			unless ( $socket->send( $packet_data, 0, $dst_sockaddr ) ) {
-				$self->_diag( $ns->[3] = "Send error: $!" );
+				$self->_diag( $ns->[3] = $self->errorstring($!) );
 				next;
 			}
 
@@ -613,8 +600,6 @@ NAMESERVER: foreach my $ns (@ns) {
 				$self->answerfrom($peerhost);
 				$self->_diag( "answer from [$peerhost]", 'length', length($buf) );
 
-				next unless $peerhost eq $ip;
-
 				my $ans = Net::DNS::Packet->new( \$buf, $self->{debug} );
 				my $error = $@;
 
@@ -682,7 +667,8 @@ sub _bgsend_tcp {
 		}
 
 		my $expire = time() + $self->{tcp_timeout} || 0;
-		return IO::Select->new( [$socket, $expire, $ns, $packet->header->id] );
+		${*$socket}{net_dns_bg} = [$expire, $ns, $packet->header->id];
+		return $socket;
 	}
 
 	$self->_diag( $self->errorstring );
@@ -713,7 +699,8 @@ sub _bgsend_udp {
 		die 'Insecure dependency while running with -T switch' if _tainted($dst_sockaddr);
 
 		my $expire = time() + $self->{udp_timeout} || 0;
-		return IO::Select->new( [$socket, $expire, $ns, $packet->header->id] );
+		${*$socket}{net_dns_bg} = [$expire, $ns, $packet->header->id];
+		return $socket;
 	}
 
 	$self->_diag( $self->errorstring );
@@ -723,53 +710,54 @@ sub _bgsend_udp {
 
 sub bgisready {
 	my $self = shift;
-	my $sel = shift || croak 'undefined argument';
+	my $sock = shift || 1;
 
-	return scalar( $sel->can_read(0.0) ) || do {
+	return scalar( IO::Select->new($sock)->can_read(0.0) ) || do {
 		return 0 unless $self->{udp_timeout};
-		my ($handle) = $sel->handles;
-		my ( $x, $expire ) = @$handle;
-		time() > $expire;
+		my $appendix = ${*$sock}{net_dns_bg} || [];
+		my ($expire) = @$appendix;
+		time() > ( $expire || return 1 );
 	};
 }
 
 
 sub bgread {
 	my $self = shift;
-	my $sel = shift || croak 'undefined argument';
+	my $sock = shift || undef;
 
-	my ($handle) = $sel->handles;
-	my ( $socket, $expire, $ip, $id ) = @$handle;
+	my $time = time();
+	my $appendix = ${*$sock}{net_dns_bg} || [$time];
+	my ( $expire, $ip, $id ) = @$appendix;
 
-	my $timeout = $expire - time();
-	return undef unless $sel->can_read( $timeout > 0 ? $timeout : 0 );
+	my $select  = IO::Select->new($sock);
+	my $timeout = $expire - $time;
+	return undef unless $select->can_read( $timeout > 0 ? $timeout : 0 );
 
 	my $buffer;
 	if ( $self->{usevc} ) {
-		$buffer = _read_tcp( $socket, INT16SZ );
+		$buffer = _read_tcp( $sock, INT16SZ );
 		return undef unless length($buffer);		# failed to get length
 		my $len = unpack 'n', $buffer;
 		return undef unless $len;			# can not determine size
 
-		unless ( $sel->can_read( $self->{tcp_timeout} ) ) {
+		unless ( $select->can_read( $timeout > 0 ? $timeout : 0 ) ) {
 			$self->_diag( $self->errorstring('timeout') );
 			return undef;
 		}
 
-		$buffer = _read_tcp( $socket, $len );
-		$self->answerfrom($ip);				# $socket->peerhost unreliable
+		$buffer = _read_tcp( $sock, $len );
+		$self->answerfrom($ip);				# $sock->peerhost unreliable
 		$self->_diag("answer from [$ip]");
 
 	} else {
-		unless ( $socket->recv( $buffer, $self->_packetsz ) ) {
+		unless ( $sock->recv( $buffer, $self->_packetsz ) ) {
 			$self->errorstring($!);
 			return undef;
 		}
 
-		my $peerhost = $socket->peerhost;
-		$self->answerfrom( $socket->peerhost );
+		my $peerhost = $sock->peerhost;
+		$self->answerfrom($peerhost);
 		$self->_diag("answer from [$peerhost]");
-		return undef unless $peerhost eq $ip;
 	}
 
 
@@ -780,8 +768,7 @@ sub bgread {
 	} else {
 		my $error  = $@;
 		my $header = $ans->header;
-		my $rcode  = $header->rcode;
-		$self->errorstring( $error || $rcode );
+		$self->errorstring( $error || $header->rcode );
 
 		return undef unless $header->qr;
 		return undef unless $header->id == $id;
@@ -829,7 +816,7 @@ sub axfr {				## zone transfer
 	my $self = shift;
 
 	my $query = $self->_axfr_start(@_);
-	my $verfy = $query->sigrr;
+	my $verfy = $query ? $query->sigrr : undef;
 	my @rr;
 	my $soa;
 
@@ -997,7 +984,8 @@ sub tsig {
 		local $SIG{__DIE__};
 		require Net::DNS::RR::TSIG;
 		Net::DNS::RR::TSIG->create(@_);
-	} || croak "$@\nunable to create TSIG record";
+	} || undef;
+	croak "$@ unable to create TSIG record" if $@;
 }
 
 
@@ -1053,11 +1041,11 @@ sub _create_tcp_socket {
 		$sock = $self->{persistent}{$sock_key};
 		$self->_diag( 'using persistent socket', $sock_key ) if $sock;
 		return $sock if $sock && $sock->connected;
-		$self->_diag('socket disconnected (trying to reconnect)');
+		$self->_diag('socket disconnected (trying to connect)');
 	}
 
 	my $srcaddr = $self->{srcaddr};
-	my $srcport = $self->{srcport} || undef;
+	my $srcport = $self->{srcport};
 	my $timeout = $self->{tcp_timeout};
 
 	if ( IPv6 && _ip_is_ipv6($ns) ) {
@@ -1071,17 +1059,17 @@ sub _create_tcp_socket {
 			Proto	  => 'tcp',
 			Timeout	  => $timeout,
 			)
-				unless USE_SOCKET_INET6;
+				if USE_SOCKET_IP;
 
 		$sock = IO::Socket::INET6->new(
 			LocalAddr => $localaddr,
-			LocalPort => $srcport,
+			LocalPort => ( $srcport || undef ),
 			PeerAddr  => $ns,
 			PeerPort  => $dstport,
 			Proto	  => 'tcp',
 			Timeout	  => $timeout,
 			)
-				if USE_SOCKET_INET6;
+				unless USE_SOCKET_IP;
 	} else {
 		my $localaddr = $srcaddr =~ /[.].*[.]/ ? $srcaddr : '0.0.0.0';
 
@@ -1097,7 +1085,7 @@ sub _create_tcp_socket {
 
 		$sock = IO::Socket::INET->new(
 			LocalAddr => $localaddr,
-			LocalPort => $srcport,
+			LocalPort => ( $srcport || undef ),
 			PeerAddr  => $ns,
 			PeerPort  => $dstport,
 			Proto	  => 'tcp',
@@ -1121,7 +1109,7 @@ sub _create_udp_socket {
 	my $ns	 = shift;
 
 	my $srcaddr = $self->{srcaddr};
-	my $srcport = $self->{srcport} || undef;
+	my $srcport = $self->{srcport};
 	my $sock_key;
 	my $sock;
 
@@ -1137,15 +1125,15 @@ sub _create_udp_socket {
 			Proto	  => 'udp',
 			Type	  => SOCK_DGRAM,
 			)
-				unless USE_SOCKET_INET6;
+				if USE_SOCKET_IP;
 
 		$sock = IO::Socket::INET6->new(
 			LocalAddr => $localaddr,
-			LocalPort => $srcport,
+			LocalPort => ( $srcport || undef ),
 			Proto	  => 'udp',
 			Type	  => SOCK_DGRAM,
 			)
-				if USE_SOCKET_INET6;
+				unless USE_SOCKET_IP;
 	} else {
 		$sock_key = 'UDP/IPv4';
 		return $sock if $sock = $self->{persistent}{$sock_key};
@@ -1162,7 +1150,7 @@ sub _create_udp_socket {
 
 		$sock = IO::Socket::INET->new(
 			LocalAddr => $localaddr,
-			LocalPort => $srcport,
+			LocalPort => ( $srcport || undef ),
 			Proto	  => 'udp',
 			Type	  => SOCK_DGRAM,
 			)
@@ -1205,13 +1193,13 @@ sub _create_dst_sockaddr {		## create UDP destination sockaddr structure
 # Lightweight versions of subroutines from Net::IP module, recoded to fix RT#96812
 
 sub _ip_is_ipv4 {
-	for ( shift || return ) {
+	for (shift) {
 		return /^[0-9.]+\.[0-9]+$/;			# dotted digits
 	}
 }
 
 sub _ip_is_ipv6 {
-	for ( shift || return ) {
+	for (shift) {
 		return 1 if /^[:0-9a-f]+:[0-9a-f]*$/i;		# mixed : and hexdigits
 		return 1 if /^[:0-9a-f]+:[0-9.]+$/i;		# prefix + dotted digits
 		return /^[:0-9a-f]+:[0-9a-f]*[%].+$/i;		# RFC4007 scoped address
@@ -1263,8 +1251,8 @@ sub udppacketsize {
 my $warned;
 
 sub make_query_packet {			## historical
-	&_make_query_packet;					# uncoverable pod
-	carp 'deprecated method; see RT#37104' unless $warned++;
+	carp 'deprecated method; see RT#37104' unless $warned++;    # uncoverable pod
+	&_make_query_packet;
 }
 
 
